@@ -113,6 +113,7 @@ void SyntaxCheck::run()
                 speller=newSpeller;
                 mReplacementList=newReplacementList;
                 mFormatList=newFormatList;
+                m_nonTextGrammarFormats=m_newNonTextGrammarFormats;
 			}
 			mLtxCommandLock.unlock();
 		}
@@ -151,14 +152,26 @@ void SyntaxCheck::run()
 		newLine.dlh->lockForWrite();
 		if (newLine.ticket == newLine.dlh->getCurrentTicket()) { // discard results if text has been changed meanwhile
             newLine.dlh->setCookie(QDocumentLine::LEXER_COOKIE,QVariant::fromValue<TokenList>(tl));
+            QList<QFormatRange>grammarOverlays=newLine.dlh->getOverlaysNoLock(m_nonTextGrammarFormats);
             foreach (const Error &elem, newRanges){
                 if(!mSyntaxChecking && (elem.type!=ERR_spelling) && (elem.type!=ERR_highlight) ){
                     // skip all syntax errors
                     continue;
                 }
-                int fmt= elem.type == ERR_spelling ? SpellerUtility::spellcheckErrorFormat : syntaxErrorFormat;
-                fmt= elem.type == ERR_highlight ? elem.format : fmt;
+                int fmt= (elem.type == ERR_spelling) ? SpellerUtility::spellcheckErrorFormat : syntaxErrorFormat;
+                fmt= (elem.type == ERR_highlight) ? elem.format : fmt;
                 newLine.dlh->addOverlayNoLock(QFormatRange(elem.range.first, elem.range.second, fmt));
+                // for ERR_highlight, remove grammarErrors
+                if(m_hideNonTextGrammarErrors && elem.type==ERR_highlight){
+                    for(int i = 0; i<grammarOverlays.size();++i){
+                        const QFormatRange &range=grammarOverlays.at(i);
+                        if(range.offset>=elem.range.first && range.offset<=elem.range.second){
+                            newLine.dlh->removeOverlayNoLock(range);
+                            grammarOverlays.removeAt(i);
+                            --i;
+                        }
+                    }
+                }
             }
             // add comment hightlight if present
             if(commentStart.first>=0){
@@ -173,6 +186,17 @@ void SyntaxCheck::run()
 			//if excessCols has changed the subsequent lines need to be rechecked.
             // don't on initial check
             if (cookieChanged) {
+                // handle runAway arguments
+                for (int i = 0; i < activeEnv.size(); i++) {
+                    if (activeEnv[i].runAway == 0) {
+                        activeEnv.remove(i);
+                        --i;
+                    }else{
+                        if(activeEnv[i].runAway>0){
+                            activeEnv[i].runAway = activeEnv[i].runAway - 1;
+                        }
+                    }
+                }
 				QVariant env;
 				env.setValue(activeEnv);
 				newLine.dlh->setCookie(QDocumentLine::STACK_ENVIRONMENT_COOKIE, env);
@@ -291,6 +315,25 @@ void SyntaxCheck::enableSyntaxCheck(const bool enable){
     if (stopped) return;
     mSyntaxChecking=enable;
 }
+/*! \brief hide non text spelling errors
+ * \param hide
+ */
+void SyntaxCheck::setHideNonTextGrammarErrors(const bool hide)
+{
+    m_hideNonTextGrammarErrors=hide;
+}
+/*!
+ * \brief SyntaxCheck::setNonTextGrammarFormats
+ * \param formats
+ */
+void SyntaxCheck::setNonTextGrammarFormats(const QList<int> formats)
+{
+    if (stopped) return;
+    mLtxCommandLock.lock();
+    newLtxCommandsAvailable = true;
+    m_newNonTextGrammarFormats=formats;
+    mLtxCommandLock.unlock();
+}
 /*!
  * \brief set character/text replacementList for spell checking
  * \param replacementList Map for characater/text replacement prior to spellchecking words. E.g. "u -> ü when german is activated
@@ -578,15 +621,21 @@ void SyntaxCheck::checkLine(const QString &line, Ranges &newRanges, StackEnviron
         if(!activeEnv.isEmpty() && activeEnv.top().endingColumn>=0 && tk.start>activeEnv.top().endingColumn){
             Environment env=activeEnv.pop();
         }
-        // special treatment for ExplSyntaxOff
-        // it is used to toggle expl3 mode off
-        if(!activeEnv.isEmpty() && activeEnv.top().name == "expl3"){
+        // handle single command env stop e.g. \ExplSyntaxOff
+        if(tk.type==Token::command){
             const QString word=tk.getText();
-            if( word == "\\ExplSyntaxOff"){
-                activeEnv.pop();
-                continue;
+            if(ltxCommands->possibleCommands["%endEnv"].contains(word)){
+                const QString envName=ltxCommands->environmentAliases.value(word);
+                if(activeEnv.top().name == envName){
+                    activeEnv.pop();
+                    continue;
+                }
             }
-            if(tk.type==Token::commandUnknown || tk.type==Token::command){
+        }
+
+        if(!activeEnv.isEmpty() && activeEnv.top().name == "%expl3"){
+            // special treatment for expl3 commands in expl3 env
+            if((tk.type==Token::commandUnknown || tk.type==Token::command)&&tk.getText()!="\\\\"){ // special treatment for \\ , see #3877
                 // collect next parts
                 // e.g. \cs_new:Npn is split into \cs _ new : Npn
                 const int start = tk.start;
@@ -724,7 +773,8 @@ void SyntaxCheck::checkLine(const QString &line, Ranges &newRanges, StackEnviron
                 // invalidates math env as active
                 Environment env;
                 env.name = "text";
-                env.id = 1; // to be changed
+                env.id = 1;
+                env.runAway = mRUNAWAYLIMIT;
                 env.dlh = dlh;
                 env.ticket = ticket;
                 env.level = tk.level;
@@ -773,7 +823,7 @@ void SyntaxCheck::checkLine(const QString &line, Ranges &newRanges, StackEnviron
                 }
             }
             word = latexToPlainWordwithReplacementList(word, mReplacementList); //remove special chars
-            if (speller->hideNonTextSpellingErrors && (checkMathEnvActive(activeEnv)||containsEnv("picture", activeEnv)) ){
+            if (speller->hideNonTextSpellingErrors && (checkMathEnvActive(activeEnv)||containsEnv("picture", activeEnv)||containsEnv("pictureHighlight", activeEnv)) ){
                 word.clear();
                 tk.ignoreSpelling=true;
             }else{
@@ -994,25 +1044,11 @@ void SyntaxCheck::checkLine(const QString &line, Ranges &newRanges, StackEnviron
 					if(env=="tikztimingtable"){
 						option="ll"; // is always 2 columns
 					}else{
-						for (int k = i + 1; k < tl.length(); k++) {
-							Token elem = tl.at(k);
-							if (elem.level < tk.level)
-								break;
-							if (elem.level > tk.level)
-								continue;
-							if (elem.subtype == Token::colDef) {
-								option = line.mid(elem.start + 1, elem.length - 2); // strip {}
-								break;
-							}
-						}
+                        option = Parsing::getArg(tl.mid(i+1),Token::colDef);
 					}
 				}
                 if(option.contains("colspec")){
-                    const QRegularExpression re{"^(.*colspec\\s*[=]\\s*\\{)(.*)\\}"};
-                    const QRegularExpressionMatch match = re.match(option);
-                    if (match.hasMatch()) {
-                        option = match.captured(2);
-                    }
+                    option=LatexTables::handleColSpec(option);
                 }
 				QSet<QString> translationMap=ltxCommands->possibleCommands.value("%columntypes");
 				QStringList res = LatexTables::splitColDef(option);
@@ -1056,9 +1092,10 @@ void SyntaxCheck::checkLine(const QString &line, Ranges &newRanges, StackEnviron
             // special treatment for \ExplSyntaxOn, \ExplSyntaxOff
             // \ProvidesExplPackage, \ProvidesExplClass and \ProvidesExplFile
             // activate latex3 mode which ignores _ in commandnames
-            if(word=="\\ExplSyntaxOn" || word=="\\ProvidesExplPackage" || word=="\\ProvidesExplClass" || word=="\\ProvidesExplFile"){
+            if(ltxCommands->possibleCommands["%beginEnv"].contains(word)){
+                const QString envName=ltxCommands->environmentAliases.value(word);
                 Environment env;
-                env.name = "expl3";
+                env.name = envName;
                 env.id = 1; // to be changed
                 env.dlh = dlh;
                 env.ticket = ticket;
@@ -1142,7 +1179,22 @@ void SyntaxCheck::checkLine(const QString &line, Ranges &newRanges, StackEnviron
                     }
 					continue;
 				}
-
+                // special treatment { \\ } in tblr (multirow cell)
+                if(word=="\\\\" && activeEnv.top().name=="tblr"){
+                    // check if this token lies with braces/none
+                    bool skipToken=false;
+                    for(int j=i-1;j>=0;--j){
+                        Token tk2=tl.at(j);
+                        if(tk2.type==Token::braces && tk2.subtype==Token::none && tk2.start+tk2.length>tk.start){
+                            // inside braces, ignore
+                            skipToken=true;
+                            break;
+                        }
+                    }
+                    if(skipToken){
+                        continue;
+                    }
+                }
 				if ((word == "\\\\") || (word == "\\tabularnewline")) {
 					if (activeEnv.top().excessCol < (activeEnv.top().id - 1)) {
 						Error elem;
@@ -1383,6 +1435,9 @@ void SyntaxCheck::checkLine(const QString &line, Ranges &newRanges, StackEnviron
 					if(options.startsWith("#")){
 						continue; // ignore type keys, like width#L
 					}
+                    if(options.endsWith("#c")){
+                        continue; // ignore values for syntax checking (#c)
+                    }
                     if(options.startsWith("%")){
                         if (!ltxCommands->possibleCommands[options].contains(word)) {
                             // special treatement for %color (mix)
